package cn.ninegame.genericframework.tools.dynamicload;


import android.content.Context;
import android.content.SharedPreferences;
import android.os.Build;
import android.text.TextUtils;
import android.util.Log;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.Closeable;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import cn.ninegame.genericframework.exception.ErrorMessage;
import cn.ninegame.genericframework.tools.ReflectHelper;
import cn.ninegame.genericframework.tools.endecode.EndecodeUtil;
import cn.ninegame.genericframework.tools.endecode.Md5Utils;
import dalvik.system.DexFile;
import dalvik.system.PathClassLoader;

/**
 * Created by lingcong on 2015/6/12.
 * Dex文件动态加载器
 */
public final class DexLoader {
    static final String TAG = "DexLoader";

    private static final int MIN_SDK_VERSION = 4;

    private static final String DEX_CLASSES_FILE = "classes.dex";
    private static final String DEX_OPT_DIR = "code_cache";

    private static final String DOT_SUFFIX = ".";
    private static final String DEX_SUFFIX = ".dex";
    private static final String JAR_SUFFIX = ".jar";
    private static final String ZIP_SUFFIX = ".zip";
    private static final String APK_SUFFIX = ".apk";
    private static final String SO_SUFFIX = ".so";

    /**
     *
     * @param context Android 的 Context
     * @param dexApk    包含Dex文件的APK包
     * @param extractDir    解压Dex文件的目录
     * @param forceReload   强制全新加载Dex
     * @throws RuntimeException
     */
    public static void loadDex(Context context,File dexApk,File extractDir,boolean forceReload) throws Exception {
        Log.i(TAG, "loadDex");

        if (Build.VERSION.SDK_INT < MIN_SDK_VERSION) {
            throw new RuntimeException(new StringBuilder().append("Multi dex installation failed. SDK ").append(Build.VERSION.SDK_INT).append(" is unsupported. Min SDK version is ").append(4).append(".").toString());
        }

            if(context == null || dexApk == null) {
                return;
            }

            if (!dexApk.exists()) {
                return;
            }

            ClassLoader loader;
            try {
                loader = context.getClassLoader();
            }
            catch (RuntimeException e) {
                Log.e(TAG, "context getClassLoader fail!");
                throw e;
            }
            if (loader == null) {
                return;
            }

            File dexFile = new File(extractDir, DEX_CLASSES_FILE);

            File dexoptDir = new File(extractDir,DEX_OPT_DIR);
            dexoptDir.mkdirs();
            File odexFile = new File(dexoptDir,DEX_CLASSES_FILE);

            if (forceReload) {
                if (dexFile.exists()) {
                    dexFile.delete();
                }

                if (odexFile.exists()) {
                    odexFile.delete();
                }
            }

            boolean bFirstLoadDex = true;
            if (odexFile.exists()) {
                bFirstLoadDex = false;
            }

            if (bFirstLoadDex) {
                //首次加载，从APK压缩包中解出DEX文件
                boolean extractSuccess = extractDexFile(context,dexApk,dexFile);
                if (!extractSuccess) {
                    return;
                }else{
                    storeInfo(context, dexFile);
                }
            }else {
                //非首次加载
                //校验DEX 的有效性，防止DEX被人篡改
                if (isModified(context,dexFile)) {
                    boolean extractSuccess = extractDexFile(context,dexApk,dexFile);
                    if (!extractSuccess) {
                        return;
                    }else{
                        storeInfo(context, dexFile);
                    }
                }
            }

            List<File> files = new ArrayList<File>();
            files.add(dexFile);

            //加载DEX
            installDexes(loader, dexoptDir, files);
    }

    private static boolean extractDexFile(Context context,File apk,File dexFile) throws IOException{
        boolean bSuccess = false;
        try {
            if (dexFile.exists()) {
                dexFile.delete();
            }
            ZipFile zipFile = new ZipFile(apk);
            ZipEntry zipEntry = zipFile.getEntry(DEX_CLASSES_FILE);
            extract(zipFile, zipEntry, dexFile);
            zipFile.close();
            long dexCostTime = System.currentTimeMillis();
            Log.d("DynamicLibLoader", "extract dex storeInfo cost time " + (System.currentTimeMillis() - dexCostTime) + "ms");
            bSuccess = true;
        } catch (IOException e) {
            Log.e(TAG, ErrorMessage.EXTRACT_DEX_FILE_FAIL);
            throw new IOException(ErrorMessage.EXTRACT_DEX_FILE_FAIL, e);
        }

        return bSuccess;
    }

    private static String getKey(File file) {
        String filePath = file.getAbsolutePath();
        //MD5(文件的绝对路径)作为key值
        String key = Md5Utils.getMD5(filePath);
        return key;
    }

    private static void extract(ZipFile apk, ZipEntry dexFile, File extractTo)
            throws IOException
    {
        BufferedInputStream in = new BufferedInputStream(apk.getInputStream(dexFile));
        BufferedOutputStream out = null;

        try {
            out = new BufferedOutputStream(new FileOutputStream(extractTo));
            byte[] buffer = new byte[4096];
            int length = in.read(buffer);
            while (length != -1) {
                out.write(buffer, 0, length);
                length = in.read(buffer);
            }
        }catch (IOException e){
            Log.e(TAG, ErrorMessage.EXTRACT_FAIL);
            throw  new IOException(ErrorMessage.EXTRACT_FAIL, e);
        }
        finally {
            closeQuietly(out);
            closeQuietly(in);
        }
    }

    private static void closeQuietly(Closeable closeable) {
        try {
            if (closeable != null) {
                closeable.close();
            }
        } catch (IOException e) {
            Log.w(TAG, "Failed to close resource", e);
        }
    }

    private static boolean isModified(Context context,File file) {
        boolean bModified = true;

        if (!file.exists()) {
            return bModified;
        }

        String key = getKey(file);//file.lastModified();

        long lastTime = context.getSharedPreferences(context.getPackageName(), (Build.VERSION.SDK_INT < 11) ? Context.MODE_PRIVATE : Context.MODE_MULTI_PROCESS).getLong(key, 0);

        if (lastTime == 0) {
            return  bModified;
        }else if(lastTime == file.lastModified()){
            bModified = false;
        }

        return bModified;
    }

    private static void storeInfo(Context context,File file) {

        if (file == null) {
            return;
        }

        String key = getKey(file);
        long lastTime = file.lastModified();

        SharedPreferences sp = context.getSharedPreferences(context.getPackageName(), (Build.VERSION.SDK_INT < 11) ? Context.MODE_PRIVATE : Context.MODE_MULTI_PROCESS);
        if(sp != null) {
            SharedPreferences.Editor editor = sp.edit();
            editor.putLong(key, lastTime).commit();
        }
    }

    private static String getFileSignature(File file) {
        //计算文件的signature: base64(M8(MD5(file)+file.lastmodify))
        String fileSignature = null;
        String fileMD5 = Md5Utils.getMD5(file);
        if(fileMD5 != null) {
            fileSignature = EndecodeUtil.m8AndBase64EncodeString(fileMD5 + file.lastModified());
        }

        return fileSignature;
    }

    public static synchronized void installDexes(ClassLoader loader, File dexDir, List<File> files)
            throws Exception {
        try {
            if (!files.isEmpty()) {
                if(ALiYunOS.isAliyunOS()) {
                    ALiYunOS.install(loader, files, dexDir);
                }else if (Build.VERSION.SDK_INT >= 24) {
                    V24.install(loader,files,dexDir);
                } else if (Build.VERSION.SDK_INT >= 23) {
                    V23.install(loader, files, dexDir);
                } else if (Build.VERSION.SDK_INT >= 19) {
                    V19.install(loader, files, dexDir);
                } else if (Build.VERSION.SDK_INT >= 14) {
                    V14.install(loader, files, dexDir);
                } else {
                    V4.install(loader, files, dexDir);
                }
            }

            if (Build.VERSION.SDK_INT >= 14 && Build.VERSION.SDK_INT < 24) {
                // 计算odex 的文件名,通过odex文件是否存在判断动态加载是否成功。
                StringBuffer stringBuffer = new StringBuffer(files.get(0).getName());
                int lastIndex = stringBuffer.lastIndexOf(DOT_SUFFIX);
                stringBuffer.replace(lastIndex, stringBuffer.length() + 1, DEX_SUFFIX);

                File odexFile = new File(dexDir, stringBuffer.toString());
                if (!odexFile.exists()) {
                    throw new IOException();
                }
            }
        }catch (IOException ioe){
            Log.e(TAG, ErrorMessage.INSTALL_DEXES_FAIL);
            throw new IOException(ErrorMessage.INSTALL_DEXES_FAIL, ioe);
        }
    }

    private static void expandFieldArray(Object instance, String fieldName, Object[] extraElements)
            throws NoSuchFieldException, IllegalArgumentException, IllegalAccessException {
        Field jlrField = ReflectHelper.findField(instance, fieldName);
        Object[] original = (Object[])jlrField.get(instance);
        Object[] combined = (Object[])Array.newInstance(original.getClass().getComponentType(), original.length + extraElements.length);

        System.arraycopy(original, 0, combined, 0, original.length);
        System.arraycopy(extraElements, 0, combined, original.length, extraElements.length);
        jlrField.set(instance, combined);
    }

    private static final class V4 {
        private static void install(ClassLoader loader, List<File> additionalClassPathEntries,File optimizedDirectory)
                throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException, IOException {
            int extraSize = additionalClassPathEntries.size();

            Field pathField = ReflectHelper.findField(loader, "path");

            StringBuilder path = new StringBuilder((String)pathField.get(loader));
            String[] extraPaths = new String[extraSize];
            File[] extraFiles = new File[extraSize];
            ZipFile[] extraZips = new ZipFile[extraSize];
            DexFile[] extraDexs = new DexFile[extraSize];

            ListIterator iterator = additionalClassPathEntries.listIterator();
            while (iterator.hasNext()) {
                File additionalEntry = (File)iterator.next();
                String entryPath = additionalEntry.getAbsolutePath();
                path.append(':').append(entryPath);
                int index = iterator.previousIndex();
                extraPaths[index] = entryPath;
                extraFiles[index] = additionalEntry;
                extraZips[index] = new ZipFile(additionalEntry);

                String apkName = additionalEntry.getName();
                String odexName = TextUtils.substring( apkName, 0, apkName.lastIndexOf('.'));
                String odexFilePath =  new StringBuilder().append(optimizedDirectory.getAbsolutePath()).append(File.separator).append(odexName).append(".dex").toString();
                File odexFile = new File(odexFilePath);
                if(!odexFile.exists()){
                    odexFile.createNewFile();
                }
                extraDexs[index] = DexFile.loadDex(entryPath, odexFilePath, 0);
            }

            pathField.set(loader, path.toString());
            DexLoader.expandFieldArray(loader, "mPaths", extraPaths);
            DexLoader.expandFieldArray(loader, "mFiles", extraFiles);
            DexLoader.expandFieldArray(loader, "mZips", extraZips);
            DexLoader.expandFieldArray(loader, "mDexs", extraDexs);
        }

        /**
         * Converts a dex/jar file path and an output directory to an
         * output file path for an associated optimized dex file.
         */
        private static String optimizedPathFor(File path, File optimizedDirectory) {
        /*
         * Get the filename component of the path, and replace the
         * suffix with ".dex" if that's not already the suffix.
         *
         * We don't want to use ".odex", because the build system uses
         * that for files that are paired with resource-only jar
         * files. If the VM can assume that there's no classes.dex in
         * the matching jar, it doesn't need to open the jar to check
         * for updated dependencies, providing a slight performance
         * boost at startup. The use of ".dex" here matches the use on
         * files in /data/dalvik-cache.
         */
            String fileName = path.getName();
            if (!fileName.endsWith(DEX_SUFFIX)) {
                int lastDot = fileName.lastIndexOf(".");
                if (lastDot < 0) {
                    fileName += DEX_SUFFIX;
                } else {
                    StringBuilder sb = new StringBuilder(lastDot + 4);
                    sb.append(fileName, 0, lastDot);
                    sb.append(DEX_SUFFIX);
                    fileName = sb.toString();
                }
            }

            File result = new File(optimizedDirectory, fileName);
            return result.getPath();
        }

    }

    private static final class V14 {
        private static void install(ClassLoader loader, List<File> additionalClassPathEntries, File optimizedDirectory)
                throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException, InvocationTargetException, NoSuchMethodException,IOException{
            Field pathListField = ReflectHelper.findField(loader, "pathList");
            Object dexPathList = pathListField.get(loader);
            Object[] elements = makeDexElements(dexPathList, new ArrayList(additionalClassPathEntries), optimizedDirectory);
            if (elements.length == 0) {
                throw new IOException();
            }
            DexLoader.expandFieldArray(dexPathList, "dexElements", elements);
        }

        /**
         * Makes an array of dex/resource path elements, one per element of
         * the given array.
         */
        private static Object[] makeDexElements(Object dexPathList, ArrayList<File> files, File optimizedDirectory)
                throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
            ArrayList<Object> elements = new ArrayList<Object>();
            Method loadDexFile = ReflectHelper.findMethod(dexPathList, "loadDexFile", new Class[]{File.class, File.class});
        /*
         * Open all files and load the (direct or contained) dex files
         * up front.
         */
            for (File file : files) {
                File zip = null;
                DexFile dex = null;
                String name = file.getName();

                if (name.endsWith(DEX_SUFFIX)) {
                    // Raw dex file (not inside a zip/jar).
                    dex = (DexFile) loadDexFile.invoke(dexPathList, new Object[]{file, optimizedDirectory});
                } else if (name.endsWith(APK_SUFFIX) || name.endsWith(JAR_SUFFIX)
                        || name.endsWith(ZIP_SUFFIX) || name.endsWith(SO_SUFFIX)) {
                    zip = file;
                    dex = (DexFile)loadDexFile.invoke(dexPathList, new Object[] { file, optimizedDirectory });
                } else if (file.isDirectory()) {
                    // We support directories for looking up resources.
                    // This is only useful for running libcore tests.
                    elements.add(ElementFactory.getElement(dexPathList.getClass().getName(), file, true, null, null));
                } else {
                    Log.w(TAG, "Unknown file type for: " + file);
                }

                if ((zip != null) || (dex != null)) {
                    elements.add(ElementFactory.getElement(dexPathList.getClass().getName(), file, false, zip, dex));
                }
            }

            return elements.toArray(new Object[elements.size()]);
        }
    }

    private static final class V19 {
        private static void install(ClassLoader loader, List<File> additionalClassPathEntries, File optimizedDirectory)
                throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException, InvocationTargetException, NoSuchMethodException,IOException {
            Field pathListField = ReflectHelper.findField(loader, "pathList");
            Object dexPathList = pathListField.get(loader);
            ArrayList suppressedExceptions = new ArrayList();
            Object[] elements = makeDexElements(dexPathList, new ArrayList(additionalClassPathEntries), optimizedDirectory, suppressedExceptions);
            if (elements.length == 0) {
                throw new IOException();
            }
            DexLoader.expandFieldArray(dexPathList, "dexElements", elements);

            if (suppressedExceptions.size() > 0) {
                Field suppressedExceptionsField = ReflectHelper.findField(dexPathList, "dexElementsSuppressedExceptions");

                IOException[] dexElementsSuppressedExceptions = (IOException[])suppressedExceptionsField.get(loader);

                if (dexElementsSuppressedExceptions == null) {
                    dexElementsSuppressedExceptions = (IOException[])suppressedExceptions.toArray(new IOException[suppressedExceptions.size()]);
                } else {
                    IOException[] combined = new IOException[suppressedExceptions.size() + dexElementsSuppressedExceptions.length];

                    suppressedExceptions.toArray(combined);
                    System.arraycopy(dexElementsSuppressedExceptions, 0, combined, suppressedExceptions.size(), dexElementsSuppressedExceptions.length);

                    dexElementsSuppressedExceptions = combined;
                }

                suppressedExceptionsField.set(dexPathList, dexElementsSuppressedExceptions);
            }
        }

        private static Object[] makeDexElements(Object dexPathList, List<File> files, File optimizedDirectory, List<IOException> suppressedExceptions)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            Method loadDexFile = ReflectHelper.findMethod(dexPathList, "loadDexFile", new Class[]{File.class, File.class});

            ArrayList<Object> elements = new ArrayList<Object>();
            for (File file : files) {
                File zip = null;
                DexFile dex = null;
                String name = file.getName();
                if (name.endsWith(DEX_SUFFIX)) {
                    // Raw dex file (not inside a zip/jar).
                    dex = (DexFile) loadDexFile.invoke(dexPathList, new Object[]{file, optimizedDirectory});
                } else if (name.endsWith(APK_SUFFIX) || name.endsWith(JAR_SUFFIX)
                        || name.endsWith(ZIP_SUFFIX) || name.endsWith(SO_SUFFIX)) {
                    zip = file;
                    dex = (DexFile) loadDexFile.invoke(dexPathList, new Object[]{file, optimizedDirectory});
                } else if (file.isDirectory()) {
                    // We support directories for looking up resources.
                    // This is only useful for running libcore tests.
                    elements.add(ElementFactory.getElement(dexPathList.getClass().getName(), file, false, null, null));
                }

                if ((zip != null) || (dex != null)) {
                    elements.add(ElementFactory.getElement(dexPathList.getClass().getName(), file, false, zip, dex));
                }
            }

            return elements.toArray(new Object[elements.size()]);
        }
    }

    private static final class V23 {
        private static void install(ClassLoader loader, List<File> additionalClassPathEntries, File optimizedDirectory)
                throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException, InvocationTargetException, NoSuchMethodException,IOException {
            Field pathListField = ReflectHelper.findField(loader, "pathList");
            Object dexPathList = pathListField.get(loader);
            List suppressedExceptions = new ArrayList();
            Object[] elements = makeDexElements(dexPathList, additionalClassPathEntries, optimizedDirectory, suppressedExceptions);
            if (elements.length == 0) {
                throw new IOException();
            }
            DexLoader.expandFieldArray(dexPathList, "dexElements", elements);

            if (suppressedExceptions.size() > 0) {
                Field suppressedExceptionsField = ReflectHelper.findField(dexPathList, "dexElementsSuppressedExceptions");

                IOException[] dexElementsSuppressedExceptions = (IOException[])suppressedExceptionsField.get(loader);

                if (dexElementsSuppressedExceptions == null) {
                    dexElementsSuppressedExceptions = (IOException[])suppressedExceptions.toArray(new IOException[suppressedExceptions.size()]);
                } else {
                    IOException[] combined = new IOException[suppressedExceptions.size() + dexElementsSuppressedExceptions.length];

                    suppressedExceptions.toArray(combined);
                    System.arraycopy(dexElementsSuppressedExceptions, 0, combined, suppressedExceptions.size(), dexElementsSuppressedExceptions.length);

                    dexElementsSuppressedExceptions = combined;
                }

                suppressedExceptionsField.set(dexPathList, dexElementsSuppressedExceptions);
            }
        }

        private static Object[] makeDexElements(Object dexPathList, List<File> files, File optimizedDirectory, List<IOException> suppressedExceptions)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            Method makeDexElements = ReflectHelper.findMethod(dexPathList, "makePathElements", new Class[]{List.class, File.class, List.class});

            return (Object[])makeDexElements.invoke(dexPathList, new Object[] { files, optimizedDirectory, suppressedExceptions });
        }
    }

    private static final class V24 {
        private static void install(ClassLoader loader, List<File> additionalClassPathEntries, File optimizedDirectory)
                throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException, InvocationTargetException, NoSuchMethodException,IOException {
            Field pathListField = ReflectHelper.findField(loader, "pathList");
            Object dexPathList = pathListField.get(loader);
            List suppressedExceptions = new ArrayList();
            Object[] elements = makeDexElements(dexPathList, additionalClassPathEntries, optimizedDirectory, suppressedExceptions,loader);
            if (elements.length == 0) {
                throw new IOException();
            }
            DexLoader.expandFieldArray(dexPathList, "dexElements", elements);

            if (suppressedExceptions.size() > 0) {
                Field suppressedExceptionsField = ReflectHelper.findField(dexPathList, "dexElementsSuppressedExceptions");

                IOException[] dexElementsSuppressedExceptions = (IOException[])suppressedExceptionsField.get(loader);

                if (dexElementsSuppressedExceptions == null) {
                    dexElementsSuppressedExceptions = (IOException[])suppressedExceptions.toArray(new IOException[suppressedExceptions.size()]);
                } else {
                    IOException[] combined = new IOException[suppressedExceptions.size() + dexElementsSuppressedExceptions.length];

                    suppressedExceptions.toArray(combined);
                    System.arraycopy(dexElementsSuppressedExceptions, 0, combined, suppressedExceptions.size(), dexElementsSuppressedExceptions.length);

                    dexElementsSuppressedExceptions = combined;
                }

                suppressedExceptionsField.set(dexPathList, dexElementsSuppressedExceptions);
            }
        }

        private static Object[] makeDexElements(Object dexPathList, List<File> files, File optimizedDirectory, List<IOException> suppressedExceptions,ClassLoader classLoader)
                throws IllegalAccessException, InvocationTargetException, NoSuchMethodException {
            Method makeDexElements = ReflectHelper.findMethod(dexPathList, "makeDexElements", new Class[]{List.class, File.class, List.class,ClassLoader.class});

            return (Object[])makeDexElements.invoke(dexPathList, new Object[] { files, optimizedDirectory, suppressedExceptions, classLoader});
        }
    }
    private static final class ALiYunOS{

        private static boolean isAliyunOS(){
            boolean isAliyunOS = false;
            try {
                Class.forName("dalvik.system.LexClassLoader");
                isAliyunOS = true;
            } catch (ClassNotFoundException e) {
            }
            return isAliyunOS;
        }


        private static void install(ClassLoader localClassLoader, List<File> additionalClassPathEntries, File optimizedDirectory)
                throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException, InvocationTargetException
                , NoSuchMethodException, IOException, InstantiationException, ClassNotFoundException {
            File libFile = additionalClassPathEntries.get(0);

            String lexFileName = libFile.getName();
            lexFileName = lexFileName.replaceAll("\\.[a-zA-Z0-9]+", ".lex");

            String lexPath = optimizedDirectory + File.separator + lexFileName;

            Class<?> classLexClassLoader = Class.forName("dalvik.system.LexClassLoader");
            Constructor<?> constructorLexClassLoader = classLexClassLoader.getConstructor(String.class, String.class, String.class, ClassLoader.class);
            Object localLexClassLoader = constructorLexClassLoader.newInstance(lexPath, optimizedDirectory, libFile.getAbsolutePath(), localClassLoader);
            Method methodLoadClass = classLexClassLoader.getMethod("loadClass", String.class);
            // in order to call "ensureInit" function. (ROM2.2.3)
            try {
                methodLoadClass.invoke(localLexClassLoader, "call.ensureInit.function");
            } catch (Exception e) {
                e.printStackTrace();
            }

            setField(
                    localClassLoader,
                    PathClassLoader.class,
                    "mPaths",
                    combineArray(
                            getField(localClassLoader, PathClassLoader.class,
                                    "mPaths"),
                            splitToArray(getField(localLexClassLoader, classLexClassLoader,
                                    "mRawDexPath"))));

            setField(
                    localClassLoader,
                    PathClassLoader.class,
                    "mFiles",
                    combineArray(
                            getField(localClassLoader, PathClassLoader.class,
                                    "mFiles"),
                            getField(localLexClassLoader, classLexClassLoader,
                                    "mFiles")));
            setField(
                    localClassLoader,
                    PathClassLoader.class,
                    "mZips",
                    combineArray(
                            getField(localClassLoader, PathClassLoader.class,
                                    "mZips"),
                            getField(localLexClassLoader, classLexClassLoader,
                                    "mZips")));
            setField(
                    localClassLoader,
                    PathClassLoader.class,
                    "mLexs",
                    combineArray(
                            getField(localClassLoader, PathClassLoader.class,
                                    "mLexs"),
                            getField(localLexClassLoader, classLexClassLoader,
                                    "mDexs")));

        }

        private static Object getField(Object obj, Class<?> cl, String field)
                throws NoSuchFieldException, IllegalArgumentException,
                IllegalAccessException {
            Field localField = cl.getDeclaredField(field);
            localField.setAccessible(true);
            return localField.get(obj);
        }

        private static void setField(Object obj, Class<?> cl, String field,
                                     Object value) throws NoSuchFieldException,
                IllegalArgumentException, IllegalAccessException {
            Field localField = cl.getDeclaredField(field);
            localField.setAccessible(true);
            Object o = localField.get(obj);
            localField.set(obj, value);
        }

        private static Object combineArray(Object arrayLhs, Object arrayRhs) {
            Class<?> localClass = arrayLhs.getClass().getComponentType();
            int i = Array.getLength(arrayLhs);
            int j = i + Array.getLength(arrayRhs);
            Object result = Array.newInstance(localClass, j);
            for (int k = 0; k < j; ++k) {
                if (k < i) {
                    Array.set(result, k, Array.get(arrayLhs, k));
                } else {
                    Array.set(result, k, Array.get(arrayRhs, k - i));
                }
            }
            return result;
        }

        private static Object appendArray(Object array, Object value) {
            Class<?> localClass = array.getClass().getComponentType();
            int i = Array.getLength(array);
            int j = i + 1;
            Object localObject = Array.newInstance(localClass, j);
            for (int k = 0; k < j; ++k) {
                if (k < i) {
                    Array.set(localObject, k, Array.get(array, k));
                } else {
                    Array.set(localObject, k, value);
                }
            }
            return localObject;
        }

        private static Object splitToArray(Object string) {
            return ((String)string).split(":");
        }
    }
}
